class SynapticML
{
	constructor()
	{
		this.iteration = 0;
		this.data = {};
		this.perceptron = null;
		this.training = false;
		this.width = 32;
		this.height = 32;
		this.learnRate = 0.01;
	}
	
	async train(runtime){
		if(!this.training){
			this.iteration = 0;
			//input [x,y] => output [r,g,b,a], hidden layers 16
			this.perceptron = new synaptic.Architect.Perceptron(2,16,4);
			this.data = await this.getImageData(runtime);
			this.training = true;
			
			//training happens based on c3 timer, every x seconds we iterate our source image 
			runtime.callFunction('startTraining');
		}
		else
		{
			this.training = false;
			runtime.callFunction('stopTraining');
		}
	}
	
	async  getImageData(runtime){
		const canvas = runtime.objects.input.getFirstInstance();
		const imageData = await canvas.getImagePixelData();
		return imageData.data;
	}
	
	iterate(runtime){
		for(let x = 0; x < this.width; x++) {
			for(let y = 0; y < this.height; y++) {
				//extract rgba value from x,y coords of source image
				const px = this.pixel(this.data, x, y);
				
				//try and predict rgba value based on x,y coord
				this.perceptron.activate([x/this.width, y/this.height]);
				//backwards propagate the predicted result with an acurate model (learn)
				this.perceptron.propagate(this.learnRate/10, px);
			}
		}
		//render new image
		this.preview(runtime);
	}
	
	//convert x, y coords from imageData to rgba color array
	pixel(data, x, y) {
		const index = y * (this.width * 4) + x * 4;
		//const index2 = ((64 * y) + x) * 4;
		const r = data[index];
		const g = data[index+1];
		const b = data[index+2];
		const a = data[index+3];

		//console.log(r, g, b);
		return [r/255, g/255, b/255, a/255];
	}
	
	preview(runtime) {
		//update iteration text object
		runtime.callFunction('setIteration', "iteration", ++this.iteration);
		//loop through every pixel and perdict what rgba value from x,y position
		for(let x = 0; x < this.width; x++) {
			for(let y = 0; y < this.height; y++) {
				const rgba = this.perceptron.activate([x/this.width, y/this.height]);
				//draw rgba value in canvas snapshot
				runtime.callFunction('draw', x, y, rgba[0], rgba[1], rgba[2], rgba[3]);
			}
		}
		//after snapshot has been populted render fill snapshot to canvas
		runtime.callFunction('render');
		
		if(this.iteration >= 10000){
			this.training = false;
			runtime.callFunction('stopTraining');
			runtime.callFunction("setButtons", true);
		}
	}
}

class BrainML
{	
	constructor()
	{
		this.config = {
			hiddenLayers: [16],
		};
		
		this.net = null;
		this.data = {};
		this.iteration = 0;
		this.training = false;
		this.width = 32;
		this.height = 32;
		this.learnRate = 0.01;
	}
	
	async  getImageData(runtime){
		const canvas = runtime.objects.input.getFirstInstance();
		const imageData = await canvas.getImagePixelData();
		return imageData.data;
	}
	
	async train(runtime){
		if(!this.training){
			this.iteration = 0;

			this.net = new brain.NeuralNetwork(this.config);
			this.data = await this.getImageData(runtime);
			this.training = true;
			
			//train based on image data (10000 iterations)		
			const trainingData = [];
		
			for(let x = 0; x < this.width; x++) {
				for(let y = 0; y < this.height; y++) {
					//extract rgba value from x,y coords of source image
					const px = this.pixel(this.data, x, y);

					trainingData.push({
						input : {x: x/this.width, y: y/this.height},
						output : px
					});	
				}
			}

			this.net.trainAsync(trainingData, {
				iterations: 10000,
				callback: (e) => { 
					//update image based on current training
					this.preview(runtime);
				},
				callbackPeriod: 10, //how many iteration must elapse to invoke callback
				learningRate: this.learnRate,
				errorThreshold: 0.005,
			}).then(() => {
				//this.preview(runtime);
				//runtime.callFunction('setIteration', "done", e.iterations); 
				this.training = false;
				runtime.callFunction("setButtons", true);
			});
		}
		else
		{
			this.training = false;
			runtime.callFunction("setButtons", true);
		}
	}
	
	async iterate(runtime){	
		//render new image
		this.preview(runtime);
	}
	
	//convert x, y coords from imageData to rgba color array
	pixel(data, x, y) {
		const index = y * (this.width * 4) + x * 4;
		//const index2 = ((64 * y) + x) * 4;
		const r = data[index];
		const g = data[index+1];
		const b = data[index+2];
		const a = data[index+3];

		debugger;
		//console.log(r, g, b);
		return {r: r/255, g:g/255, b:b/255, a:a/255};
	}
	
	preview(runtime) {
		//update iteration text object
		runtime.callFunction('setIteration', "iteration", ++this.iteration);
		
		//loop through every pixel and perdict what rgba value from x,y position
		for(let x = 0; x < this.width; x++) {
			for(let y = 0; y < this.height; y++) {
				const rgba = this.net.run({x:x/this.width, y: y/this.height});
				//draw rgba value in canvas snapshot
				runtime.callFunction('draw', x, y, rgba.r, rgba.g, rgba.b, rgba.a);
			}
		}
		//after snapshot has been populted render fill snapshot to canvas
		runtime.callFunction('render');
	}
}